#! /usr/bin/env python
#
#  Author: Tomas de la Rosa
#
#  This software is just to simplify daily research labor
#  of imagining graphs when looking for some insight.
#  Many times I find myself drawing this in a paper,
#  spending some time in get them right.

import argparse
import sys
import re
import os.path, shutil

taskvars = dict()
taskoperators = dict()

# PLANNER VERSIONS 
vFD = 0
vLAMA08 = 1
vTFD = 2
vUNKNOWN = -1
taskversion = -1

class state_var (object):
	def __init__ (self, name, vrange):
		self.name = name    # A mapping name to the groups file
		self.vrange = vrange  # Could be different from the groups domain size, due to extra values in the temporal representation

                self.init = None
		self.goal = None
		self.domain = None

class action (object):
	# for now I only need ground action name
	def __init__ (self, opstring):
		self.name = opstring.split()
		self.snaptype = None
	def __str__ (self):
		return "<Ation {%s}>:%s" % (self.snaptype, self.name)
		
		
class sas_transition (object):
	def __init__(self, target, op, transtype='c'):
		self.target = int(target)
		self.op = int(op)
		self.transtype = transtype
		self.conditions = []
		self.duration = -1
	def  __str__(self):
		ostr ="<SAS+Tranisiton: to %d, op %d, type %s, dur:%s>" % (self.target, self.op, self.transtype, self.duration)
		for c in self.conditions:
			ostr+= "\n   %s" % c
		return ostr
	

def cleanatoms(group_atom):
	a = group_atom.replace("Atom","").strip()
	a = a.replace("("," ").replace(")"," ")
	a = a.replace("<","").replace(">","")
	a = a.replace(",","")
	return a.strip()

def check_version(filepath):
	sasfile = open(filepath,'r')
	sasdata = sasfile.read()
	version = re.findall('begin_version\n(.*?)\nend_version', sasdata, re.DOTALL)

        
        if version and version[0]=='3':
                print "SAS output 3 (Current FD representation)", version
                return vFD
        else:
                print "Unknown SAS ouput version..."
                return vUNKNOWN


def read_variables(filepath):
        global taskvars

	sasfile = open(filepath,'r')
	sasdata = sasfile.read()
	varsdata = re.findall('begin_variable\n(.*?)\nend_variable', sasdata, re.DOTALL)
        taskvars = dict()

        for ivar in range(len(varsdata)):
		varlines = varsdata[ivar].split('\n')
                line = 0

                name = varlines[line]
                varrange = int(varlines[line+2])

                taskvars[ivar] = state_var(name, varrange)
                line+= 3
                vardomain = dict()
                for jvalue in range(varrange):
                        vardomain[jvalue] = cleanatoms(varlines[line+jvalue])
                taskvars[ivar].domain = vardomain

        return taskvars
                        

        

def read_variables_old(filepath, groupsfilepath):
	sasfile = open(filepath,'r')
	sasdata = sasfile.read()
	
        vardata = re.findall('begin_variables\n(.*?)\nend_variables', sasdata, re.DOTALL)[0]
	split_var_data = vardata.split('\n')[1:]
	vars_data_list = [v.split() for v in split_var_data]
		
	for var, ivardata in zip(range(len(vars_data_list)), vars_data_list):
		taskvars[var] = state_var(ivardata[0], int(ivardata[1]))
        sasfile.close()
 

        # Creating groups from the groups file. Then it will be added as the domain of taskvars
	varfile = open(groupsfilepath,'r')
	groupsdata = varfile.readlines()
        varfile.close()

	sasvardata = dict()
	
	for line in groupsdata:
		newsasvar = re.match('var([0-9]+)', line)
		if newsasvar:
			sasvar = newsasvar.group() 
			sasvardata[sasvar] = dict()
		else:
			index_value = line.split(':')
			#print ">>", sasvar, index_value
			sasvardata[sasvar][int(index_value[0].strip())]= cleanatoms(index_value[1])

	for vari in taskvars.keys():
		varname = taskvars[vari].name
                taskvars[vari].domain = sasvardata[varname]

	return taskvars 

		
def read_DTGs (filepath):
	sasfile = open(filepath,'r')
	sasdata = sasfile.read()

	dtgs = re.findall('begin_DTG\n(.*?)\nend_DTG', sasdata, re.DOTALL)
	sasfile.close()
	return dtgs


def read_task (filepath, fulltemporal=False):
	global taskoperators
	
	sasfile = open(filepath,'r')
	sasdata = sasfile.read()
	
        # for older versions of sasfile
        if taskversion != vFD:
                vardata = re.findall('begin_variables\n(.*?)\nend_variables', sasdata, re.DOTALL)[0]
	
        initdata = re.findall('begin_state\n(.*?)\nend_state', sasdata, re.DOTALL)
	goaldata = re.findall('begin_goal\n(.*?)\nend_goal', sasdata, re.DOTALL)
	
	operatordata = re.findall('begin_operator\n(.*?)\nend_operator', sasdata, re.DOTALL)
		
	if len(operatordata) == 0:
		raise Exception ("No operator data found.  Check you are using correct planning model in input files")
		
	sasfile.close()

	#split_var_data = vardata.split('\n')[1:]
	#vars_data_list = [v.split() for v in split_var_data]
	
	initial_state = initdata[0].split('\n')

	#for var, istate, ivardata in zip(range(len(initial_state)), initial_state, vars_data_list):
	#	taskvars[var] = state_var(ivardata[0], float(istate), int(ivardata[1]))
        for var, istate in zip(range(len(initial_state)), initial_state):
                taskvars[var].init = float(istate)

	goal_state = goaldata[0].split('\n')
	
	assert int(goal_state[0]) == len(goal_state[1:])
	
	for igoal in goal_state[1:]:
		vari, value = tuple(map(int, igoal.split()))
		taskvars[vari].goal = value

	for iop in range(len(operatordata)):
		oplines = operatordata[iop].split('\n')
		taskoperators[iop] = action(oplines[0])
		#if read_snapactions:
		#	taskoperators[iop].snaptype = oplines[1]
		#else:
		#	taskoperators[iop].snaptype = "c"  # compressed actions
		taskoperators[iop].snaptype = "c"
		
		
	return taskoperators


def split_tfd_dtg_blocks (dtginput, sasvar):

	vardomain = sasvar.domain
	dtgdata = dtginput.split('\n')
	
	line = 0
	transition_list = dict()
		
	for ivalue in range(sasvar.vrange):
		transition_list[ivalue] = []
		ntransitions = int(dtgdata[line])

		line += 1 
		for jtransition in range(ntransitions):

			new_t = sas_transition(*dtgdata[line: line+3])
			#print dtgdata[line+3]
			durcomp, durvar = tuple(dtgdata[line+3].split())

			# duration is represented as a numeric value of a state variable
			new_t.duration = (durcomp, taskvars[int(durvar)].init)
			line = line + 4
			nconds = int(dtgdata[line])
			#print "num conditions", nconds
			line += 1
			for kcond in range(nconds):
				new_t.conditions.append(dtgdata[line+kcond].split())
			line = line + nconds
			
			transition_list[ivalue].append(new_t)

	return transition_list
		
def split_dtg_blocks (dtginput, sasvar):

	vardomain = sasvar.domain
	dtgdata = dtginput.split('\n')
	
	line = 0
	transition_list = dict()
		
	for ivalue in range(sasvar.vrange):
		transition_list[ivalue] = []
		ntransitions = int(dtgdata[line])
		#print " >> n trasitions ", ntransitions
		line += 1 
		for jtransition in range(ntransitions):
			
			new_t = sas_transition(dtgdata[line], dtgdata[line+1])
			#print "  >> trans", dtgdata[line:line+2]
			line = line + 2
			nconds = int(dtgdata[line])
			#print "  >>> nconds", nconds

			line += 1
			for kcond in range(nconds):
				new_t.conditions.append(dtgdata[line+kcond].split())
			line = line + nconds
			
			transition_list[ivalue].append(new_t)

	return transition_list

	
def read_CG (filepath):
	sasfile = open(filepath,'r')
	#sasdata = sasfile.read()
	#cg_data = re.findall('begin_CG\n(.*?)\nend_CG', sasdata, re.DOTALL)[0]
	sasdata = sasfile.readlines()
	ingraph = False

	cg_data = []
	for line in sasdata:
		
		if ingraph:
			if line.strip() == "end_CG":
				ingraph = False
			else:
				#print line.strip()
				target_weigth = line.strip().split()
				cg_data.append(int(target_weigth[0]))
			
		elif line.strip() == "begin_CG":
			ingraph = True

	sasfile.close()

	return cg_data

def split_blocks(data):
	i = 0
	blocks = []
	while (i < len(data)):
		blocks.append(data[i+1:i+1+data[i]])
		i=i+1+cg_data[i]
	return blocks


def maptransvalue (value, domain):
	if value in domain.keys():
		return domain[value]
	else:
		return "!<%d>" % value

	
def get_dtg_dot (var, transitions, args):
	values = var.domain
	
	dot = "digraph G { \n"
	for ivalue in transitions.keys():
		if (var.goal == ivalue):
			nodefeatures = ",style=filled, fillcolor=palegreen"
		elif (var.init == ivalue):
			nodefeatures = ",style=filled, fillcolor=lightblue"
		else:
			nodefeatures = ""
			
		dot += 'v%d [label="%s"%s]\n' % (ivalue, maptransvalue(ivalue, values), nodefeatures)

	for ivalue in transitions.keys():
		for jtrans in transitions[ivalue]:
			if (taskoperators[jtrans.op].snaptype == 'e' or taskoperators[jtrans.op].snaptype == 's' or args.temporal):
				opname = taskoperators[jtrans.op].name[0] + ' ' + taskoperators[jtrans.op].name[-1]
			else:
				opname = taskoperators[jtrans.op].name[0]
			if (not args.showconditions):
				if args.durativeactions:
					dot += 'v%d -> v%d [label="%s\n%s"]\n' % (ivalue, jtrans.target, opname, "".join(map(str,jtrans.duration)))
				else:
					dot += 'v%d -> v%d [label="%s"]\n' % (ivalue, jtrans.target, opname)
				
			else:
				cond_str = ""
				for cond in jtrans.conditions:
					cond_str += "{%s}var%s=%s\n" % (cond[2],cond[0],cond[1])
				if args.durativeactions:
					dot += 'v%d -> v%d [fontsize="10" label="%s\n%s\n%s"]\n' % (ivalue, jtrans.target, opname, "".join(map(str,jtrans.duration)), cond_str)
				else:
					dot += 'v%d -> v%d [fontsize="10" label="%s\n%s"]\n' % (ivalue, jtrans.target, opname, cond_str)
				
				
	dot += "}"
	return dot

def get_cg_dot (blocks):
	dot = "digraph G { \n"
	for var in range(len(blocks)):
		for target in blocks[var]:
			dot += "%s -> %s \n" % (taskvars[var].name, taskvars[target].name)
	dot += "}"
	return dot


def create_parser ():
    argparser = argparse.ArgumentParser (description="Finite Domain Representation to Graphviz")

    mandatory = argparser.add_argument_group ("Mandatory arguments", "The following arguments are required.")
    mandatory.add_argument ('-f', '--sasfile',
                            required=True,
                            help="The output file from the FD pre-proccessor")
    mandatory.add_argument ('-o', '--outputdir',
			    required=True,
			    help="Destination folder that will be created with the graphs")
    
    
    
    optional = argparser.add_argument_group ('Optional')
    
    optional.add_argument ('-V', '--plannerversion',
			   help= "For specifying old planner version such as lama2008 or tfd")

    optional.add_argument ('-g', '--groupsfile',
                            help="The test.groups or variable.groups file describing SAS var equivalence")

    optional.add_argument ('-D', '--durativeactions',
			   default=False,
			   action='store_true',
			   help= "Read durative action representation from TFD ")

    optional.add_argument ('-C', '--showconditions',
			   default=False,
			   action='store_true',
			   help= "Show the transition conditions in DTG graphs ")

    optional.add_argument ('-T', '--temporal',
			   default=False,
			   action='store_true',
			   help= "Use the full temporal domain representation with snap actions ")
		
    optional.add_argument ('-O', '--outputname',
			   default='fdr',
			   help="output name for files")
        
    # and return the parser
    return argparser



if __name__ == "__main__":

	print " Graphing Causal Graph and DTG from Fast-Downward ouput \n"
	
	argparser = create_parser()
	args = argparser.parse_args()

	if os.path.exists(args.outputdir):
		shutil.rmtree(args.outputdir)
		
	os.mkdir(args.outputdir)
	os.mkdir(os.path.join(args.outputdir, "dotfiles"))
	print "Copying source files..."
	fname = os.path.basename(args.sasfile)
	shutil.copyfile(args.sasfile, os.path.join(args.outputdir, fname))
        
        if args.groupsfile:
                fname = os.path.basename(args.groupsfile)
                shutil.copyfile(args.groupsfile, os.path.join(args.outputdir, fname))

        
	if args.plannerversion:
                if args.plannerversion.upper() == 'LAMA2008':
                        taskversion = vLAMA08
                elif args.plannerversion.upper() == 'TFD':
                        taskversion = vTFD
                elif args.plannerversion.upper() == 'FD':
                        taskversion = vFD

        else:
                taskversion = check_version(args.sasfile)

        if taskversion == vUNKNOWN:
                print "You should use the current FD output format or indicate an implemented planner version, eg., LAMA2008, TFD"
                sys.exit(-1)

	print "Reading Task..."

	cg_data = read_CG(args.sasfile)
	cg_blocks = split_blocks(cg_data)

       
        #sys.exit()

        if taskversion == vFD:
                taskvars = read_variables(args.sasfile)
        elif taskversion == vLAMA08:
                if not args.groupsfile:
                        print " This output version needs the groupsfile 'test.groups' to process atoms."
                        sys.exit(-1)

                taskvars = read_variables_old(args.sasfile, args.groupsfile)

        elif taskversion == vLAMA08 or taskversion == vTFD:
                if not args.groupsfile:
                        print " This output version needs the groupsfile 'variable.groups' to process atoms"
                        sys.exit(-1)
                taskvars = read_variables_old(args.sasfile, args.groupsfile)

	sastransitions = dict()
	taskoperators = read_task (args.sasfile, args.temporal)

	print "Writing the Causal Graph...\n"
	dotfile = os.path.join(args.outputdir, args.outputname+"_cg.dot")
	dotfile_moved = os.path.join(args.outputdir, "dotfiles", args.outputname+"_cg.dot")
	dotstream = open(dotfile,'w')
	dotstream.write( get_cg_dot(split_blocks(cg_data)) )
	dotstream.close()

	cmd = "dot -Tpdf -O " + dotfile
	os.system(cmd) 
	shutil.move(dotfile, dotfile_moved)
	
	print "Writing the Domain Transition Graphs...\n"
	
	dtgs = read_DTGs (args.sasfile)
	for vari, dtg in zip(range(len(dtgs)), dtgs):

		if -2 in taskvars[vari].domain.keys():
			continue

		print " .. for ", taskvars[vari].name

		if args.durativeactions:
			sastransitions[vari] = split_tfd_dtg_blocks(dtgs[vari],taskvars[vari])
		else:
			sastransitions[vari] = split_dtg_blocks(dtgs[vari],taskvars[vari])

		dot_str = get_dtg_dot(taskvars[vari], sastransitions[vari], args)
		dotfile = os.path.join(args.outputdir, args.outputname+"_"+taskvars[vari].name+"_dtg.dot")
		dotfile_moved = os.path.join(args.outputdir, "dotfiles", args.outputname+"_"+taskvars[vari].name+"_dtg.dot")
		dotstream = open(dotfile,'w')
		dotstream.write(dot_str)
		dotstream.close()

		cmd = "dot -Tpdf -O " + dotfile
		os.system(cmd)
		shutil.move(dotfile, dotfile_moved)

	print " Done!"
		
	
	
	

	
		
